
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef _MODEL_H_
#define _MODEL_H_

#include "classes.h"
#include "locator.h"
#include "../shared.src/color.h"
#include "texture.h"


class model : public locator {
public:
	model();
	model(const model &m);
	~model();

	model &operator=(const model &m);

	/// sets m_ambient_color
	void set_ambient_color(const color &c);
	/// sets m_diffuse_color
	void set_diffuse_color(const color &c);
	/// sets m_specular_color
	void set_specular_color(const color &c);

	/// returns m_ambient_color
	inline const color &get_ambient_color() const {return m_ambient_color;}
	/// returns m_diffuse_color
	inline const color &get_diffuse_color() const {return m_diffuse_color;}
	/// returns m_specular_color
	inline const color &get_specular_color() const {return m_specular_color;}

	/// Sets color of 'ambient', 'diffuse' or 'specular'
	void set_color(const std::string colorname, const color &c);

	/// sets m_bumpyness
	void set_bumpyness(GAIM_FLOAT b);
	/// returns m_bumpyness
	inline GAIM_FLOAT get_bumpyness() const {return m_bumpyness;}

	/// sets m_shininess
	void set_shininess(GAIM_FLOAT s);
	/// returns m_shininess
	inline GAIM_FLOAT get_shininess() const {return m_shininess;}

	/// sets m_direct_lighting_factor
	void set_direct_lighting_factor(GAIM_FLOAT dlf);
	/// sets m_reflection_factor
	void set_reflection_factor(GAIM_FLOAT rf);
	/// sets m_refraction_factor
	void set_refraction_factor(GAIM_FLOAT rf);
	/// sets m_refractive_index
	void set_refractive_index(GAIM_FLOAT ri);

	/// returns m_direct_lighting_factor
	inline GAIM_FLOAT get_direct_lighting_factor() const {return m_direct_lighting_factor;}
	/// returns m_reflection_factor
	inline GAIM_FLOAT get_reflection_factor() const {return m_reflection_factor;}
	/// returns m_refraction_factor
	inline GAIM_FLOAT get_refraction_factor() const {return m_refraction_factor;}
	/// returns m_refractive_index
	inline GAIM_FLOAT get_refractive_index() const {return m_refractive_index;}

	/// sets m_mesh
	void set_mesh(const mesh *m);
	/// returns m_mesh
	inline const mesh *get_mesh() const {return m_mesh;}

	/// renders this model using OpenGL
	void to_OpenGL() const;

	/// Returns m_ambient_texture
	inline const texture &get_ambient_texture() const {return m_ambient_texture;}
	/// Returns m_diffuseTexture
	inline const texture &get_diffuse_texture() const {return m_diffuse_texture;}
	/// Returns m_specularTexture
	inline const texture &get_specular_texture() const {return m_specular_texture;}
	/// Returns m_shininessTexture
	inline const texture &get_shininess_texture() const {return m_shininess_texture;}
	/// Returns m_bump_map_texture
	inline const texture &get_bump_map_texture() const {return m_bump_map_texture;}

	/// Returns m_ambient_texture
	inline texture &get_ambient_texture() {return m_ambient_texture;}
	/// Returns m_diffuseTexture
	inline texture &get_diffuse_texture() {return m_diffuse_texture;}
	/// Returns m_specularTexture
	inline texture &get_specular_texture() {return m_specular_texture;}
	/// Returns m_shininessTexture
	inline texture &get_shininess_texture() {return m_shininess_texture;}
	/// Returns m_bump_map_texture
	inline texture &get_bump_map_texture() {return m_bump_map_texture;}

	/// Sets m_ambient_texture
	void set_ambient_texture(const std::string &filename);
	/// Sets m_diffuse_texture
	void set_diffuse_texture(const std::string &filename);
	/// Sets m_specular_texture
	void set_specular_texture(const std::string &filename);
	/// Sets m_shininess_texture
	void set_shininess_texture(const std::string &filename);
	/// Sets m_bump_map_texture
	void set_bump_map_texture(const std::string &filename);

	/**
	Sets texture of 'ambient', 'diffuse', 'specular' or 'shininess'.
	Returns true if texture loading is succesfful
	*/
	bool set_texture(const std::string colorname, const char *filename);

	/// Sets texture enabled of 'ambient', 'diffuse', 'specular' or 'shininess'.
	void set_texture_enabled(const std::string colorname, bool e);

	/// Turns this object into a string (e.g., for storing a scene to file)
	std::string to_string(const std::string &basepath) const;

	/// Returns bounding sphere (from mesh, but un-dualized), transform according to xf.
	GAIM_RETURN_TYPE get_bounding_sphere() const;

	/**
	Finds the intersection of ray \param R and this model.
	Depending on the value of \param find_closest , either a search
	for the closes intersection will be done, or for _any_ intersection.
	For shadow rays, for instance, we only need to know that they intersect
	a model, now what model or where.
	When an intersection is found, and \param find_closest is true, 
	information about that point is stored in \param spt .

	This function simply transforms \param R into the local frame and
	calls get_mesh()->find_intersection() to do the actual intersection test.
	*/
	bool find_intersection(const ray &R, surface_point &spt, bool find_closest) const;

	void compute_surface_point_details(surface_point &spt) const;

	/// Returns texture \param texname
	texture &get_texture(const std::string &texname);

	/// Returns texture \param texname
	const texture &get_texture(const std::string &texname) const;

private:
	const mesh* m_mesh; ///< the mesh of this model; object destructed by scene, not by this class

	color m_ambient_color; ///< ambient color of this model (unles ambient texture is specified)
	color m_diffuse_color; ///< diffuse color of this model (unles diffuse texture is specified)
	color m_specular_color; ///< specular color of this model (unles specular texture is specified)


	texture m_ambient_texture;
	texture m_diffuse_texture;
	texture m_specular_texture;

	GAIM_FLOAT m_shininess; ///< shininess of specular component
	texture m_shininess_texture; // todo in UI & to_string() & rsn.cpp

	GAIM_FLOAT m_bumpyness; ///< modulates the m_bump_map_texture
	texture m_bump_map_texture;

	GAIM_FLOAT m_direct_lighting_factor; // weight assigned to 'OpenGL' lighting [0 ... 1]
	GAIM_FLOAT m_reflection_factor; // weight assigned to reflections [0 ... 1]
	GAIM_FLOAT m_refraction_factor; // weight assigned to refractions [0 ... 1]

	GAIM_FLOAT m_refractive_index; ///<

	/// Sets material to simple white
	void set_default_material_colors();



};


#endif /* _MODEL_H_ */
