
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef _MODELING_WINDOW_H_
#define _MODELING_WINDOW_H_

#include <stdlib.h>

#include <vector>
#include <utility>

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include <FL/Fl_Gl_Window.H>

#include "classes.h"

class modeling_window : public Fl_Gl_Window {
public:
	modeling_window(int x, int y, int w, int h, const char *label = NULL);
	virtual ~modeling_window();

	/// Called by FLTK to draw the scene
	virtual void draw();

	/// Called by FLTK to handle UI events
	virtual int handle(int event);

private:
	/**
	Loads the current projection into OpenGL (using current camera from g_ui_state)
	Also sets the viewport
	*/
	void setup_GL_projection_matrix();

	/// returns m_glFrustum_near
	inline GLdouble glFrustum_near() const {return 0.1;}
	/// returns m_glFrustum_far
	inline GLdouble glFrustum_far() const {return 100.0;}


	/// loads the current camera positon into OpenGL (using current camera from g_ui_state)
	void setup_GL_modelview_matrix();

	/// Sets up OpenGL lighting
	void setup_GL_lighting(const scene &S);

	/// Called by handle() to handle FL_PUSH
	void handle_push();
	/// Called by handle() to handle FL_RELEASE
	void handle_release();
	/// Called by handle() to handle FL_MOVE
	void handle_move();
	/// Called by handle() to handle FL_DRAG
	void handle_drag();

	/**
	Implement cycled picking (all required info is retrieved from FLTK event state / GL state).
	Returns the pick record (object index, object z)
	*/
	std::pair<int, int> cycle_pick();

	/// Called to select object (model or light) number \param idx
	void select_object(int idx);

	/**
	TODO: draw only one object!!!
	Returns the 

	(\param y is 'inverted' (0 = bottom, h() = top))

	The implementation uses a full redraw of the scene, so cache
	result whenever possible!
	*/
	GLfloat get_world_z(int object_idx, int x, int y);

	/// returns m_mouse_mode
	mouse_modes get_mouse_mode() const;

	/// mouse position on previous drag / push (whatever was the latest event)
	c3ga m_drag_mouse_pos;

	/// returns the mouse position as reported by FLTK as a 3D euclidean vector
	c3ga get_FLTK_mouse_position() const;

	/// returns the center of the window as a 3D euclidean vector
	c3ga get_window_center() const;

	/**
	Returns true if mouse \param pos is inside the 'inner circle' of the spaceball interface.	
	Inside: rotate camera in plane orthogonal to screen plane.
	Outside: rotate camera in screen plane.
	*/
	bool inside_spaceball_circle(const c3ga &pos);

	/* START OF VARIABLES THAT ARE SET IN RESPONSE TO FL_PUSH */
	/**
	The mouse mode for the current mouse 'action'
	This variable is set every time the mouse button is pushed:
	the value comes from g_ui_state->get_mouse_mode();
	*/
	mouse_modes m_mouse_mode; 

	/// true if FL_PUSH was inside spaceball inner circle
	bool m_inside_spaceball_circle;


	/* END OF VARIABLES THAT ARE SET IN RESPONSE TO FL_PUSH */

	/**
	Used for OpenGL picking.
	When m_OpenGL_pick[0] and m_OpenGL_pick[1] are >= 0,
	setup_GL_projection_matrix() will call gluPickMatrix() to setup a picking matrix.
	*/
	int m_OpenGL_pick[4];

	/// sets m_OpenGL_pick to {-1, -1, -1, -1}
	void reset_OpenGL_pick(); 

	/// returns true if picking mode is active according to m_OpenGL_pick
	bool  OpenGL_pick_active() const;

	/**
	Implements picking.
	\param x and \param y are the window coordinates where the pick
	should occur (\param y is 'inverted' (0 = bottom, h() = top))
	\param w and \param h are the width and height of the 'picking window'		

	Returns a vector of pairs of GLints, that represent models / lights that were picked,
	are the average z-distance of the hit.

	The 'names' (ints) are organized as
	[0 ... nb_models-1] : the models
	[nb_models ... nb_models + nb_lights-1] : the lights
	*/
	std::vector<std::pair<GLuint, GLuint> > pick(int x, int y, int w, int h);

	/// The location of the previous pick 
	int m_previous_pick_x, m_previous_pick_y;
	/// The hits registered on the previous pick (used for cycling through the hits)
	std::vector<GLuint> m_previous_pick_hits;


	/**
	When >= 0, this variable causes draw() to only draw that specific object.
	This is used to speed up z-lookups
	*/
	int m_restrict_draw_to_object_idx;

	/// sets m_restrict_draw_to_object_idx to -1
	void reset_restrict_draw_to_object_idx();

	/**
	When mouse mode is MM_TRANSLATE_ORTHOGONAL or 
	MM_TRANSLATE_PARALLEL, this value is set during handle_push().
	It is the depth (z) of the selected object at the position of the mouse.
	This value is used to make the selected object 'stick' to the cursor 
	(for MM_TRANSLATE_PARALLEL).
	TODO NO uSE FOR MM_TRANSLATE_ORTHOGONAL?????
	*/
	GLfloat m_selection_depth;

	/// Returns m_selection_depth
	GLfloat get_selection_depth() const;

	/**
	The versor that is orbited when mouse_mode is MM_CAMERA_ORBIT.
	This value is set in handle_push().
	*/
	c3ga m_orbit_versor;

	/// Returns m_orbit_versor
	const c3ga &get_orbit_versor() const;

	/// Quadric for rendering spheres with GLU
	GLUquadricObj *m_sphere_quadric;

	/// The point where the user last clicked on the bounding sphere (when mouse mode is MM_ROTATE)
	c3ga m_previous_rotate_point;

	/// Returns m_previous_rotate_point
	const c3ga &get_previous_rotate_point() const;

	/**
	Computes the point used for rotation computations (when mouse mode is MM_ROTATE).
	Thev value this function returns is written to m_previous_rotate_point .
	*/
	c3ga get_rotate_point() const;


	/// returns a sphere that bounds the selected object
	c3ga get_selected_object_bounding_sphere() const;

	/// Returns the OpenGL name used during picking for the bounding sphere
	int get_bounding_sphere_GL_name() const;

	/**
	Computes a sphere (point or real sphere) at the intersection of the 
	line from camera \param C though pixel \param mouse_position 
	and the sphere \param bounding_sphere.
	(USED to throw an exception (std::string) if no intersection exists.)
	Now uses the intersection with the plane at \param bounding_sphere
	as a fallback (experimental)
	*/
	c3ga compute_intersection_location(const camera &C, 
												 const c3ga &mouse_position,
												 const c3ga &bounding_sphere) const;


	/// sets OpenGL to wireframe mode
	void enable_wireframe_mode() const;

	/// return sets OpenGL to regularmode
	void disable_wireframe_mode() const;


};


#endif /* _MODELING_WINDOW_H_ */
