
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef _PROPERTIES_WINDOW_H_
#define _PROPERTIES_WINDOW_H_

#include <string>
#include <FL/Fl_Window.H>
#include <FL/Fl_Color_Chooser.H>
#include <FL/Fl_Check_Button.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Input.H>
#include <FL/Fl_Value_Slider.H>

#include "classes.h"

class properties_window;

// properties window color chooser
class pw_color_chooser {
public:
	pw_color_chooser(properties_window *PW, bool have_color, bool have_texture, int x, int y, int w, int h, const char *color_name);
	~pw_color_chooser() {}

	void callback(Fl_Widget *W, void *Data); ///< called by properties_window callback

	/// Returns m_height
	inline int h() const {return m_height;}

	/// Updates widgets according to \param c and \param t
	void update(const color &c, const texture &t);	

	/// Updates widgets according to \param c (if no texture)
	void update(const color &c);	

private:
	Fl_Box *m_label;
	Fl_Color_Chooser *m_color_chooser;
	Fl_Check_Button *m_texture_check_button;
	Fl_Input *m_texture_filename_input;
	Fl_Button *m_texture_browse_button;
	int m_height; /// height of all widgets combined
};


/**
This window display model & light source properties that can be edited.
It has two modes, one for models, one for light sources.
Toggle by calling use set_X_active().
*/
class properties_window : public Fl_Window {
public:
	properties_window(int w, int h, const char* label=0);
	properties_window(int x, int y, int w, int h, const char* label=0);
	virtual ~properties_window();

	// called by pw_callback (in anonymous namespace)
	void callback(Fl_Widget *W, void *Data); 

	/// Called by UI when a new model has been selected (updates all widgets for new model)
	void update_selected_model(const model &m);
	
	/// Called by UI when a new light source has been selected (updates all widgets for new light source)
	void update_selected_light_source(const light_source &l);

	/// Called by UI when no object has been selected (updates widgets)
	void update_selected_none();

	/// Called by pw_color_chooser to set the color of [ambient/diffuse/specular]
	void set_color(const std::string &name, const ::color &c);

	/// Called by pw_color_chooser to set the texture of [ambient/diffuse/specular]
	bool set_texture(const std::string &what, const char *filename);

	/// Called by pw_color_chooser to set the texture enabled of [ambient/diffuse/specular]
	void set_texture_enabled(const std::string &what, bool e);


private:
	void setup_widgets(); /// Creates all the widgets


	Fl_Group *m_model_group; ///< the group containing the modeling widgets
	Fl_Group *m_light_source_group; ///< the group containing the light source widgets

	/*
	The widgets of the model group:
	*/

	/// ambient color (model):
	pw_color_chooser *m_model_ambient_color_chooser;
	/// diffuse color (model):
	pw_color_chooser *m_model_diffuse_color_chooser;
	/// specular color (model):
	pw_color_chooser *m_model_specular_color_chooser;

	// shininess 
	Fl_Value_Slider *m_shininess_slider;
	pw_color_chooser *m_shininess_color_chooser;

	// bumpmap
	Fl_Value_Slider *m_bumpyness_slider;
	pw_color_chooser *m_bump_map_color_chooser; ///actually, used for just the  text input field . . 

	Fl_Value_Slider *m_direct_lighting_slider;
	Fl_Value_Slider *m_reflection_slider;
	Fl_Value_Slider *m_refraction_slider;
	Fl_Value_Slider *m_refractive_index_slider;

	/*
	The widgets of the light source group:
	*/

	/// ambient color (light source):
	pw_color_chooser *m_light_source_ambient_color_chooser;
	/// diffuse color (light source):
	pw_color_chooser *m_light_source_diffuse_color_chooser;
	/// specular color (light source):
	pw_color_chooser *m_light_source_specular_color_chooser;

	Fl_Value_Slider *m_spot_exponent_slider;

	/// set the label of the window
	void set_label(const std::string &pw_label_str);

};

#endif /* _PROPERTIES_WINDOW_H_ */
