
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <string>

// disable some annoying warnings in VC++ due to FLTK (1.1.6)
#ifdef WIN32
#pragma warning( disable : 4311 4312 4244 )
#endif

#include <FL/Fl.h>
#include <FL/Fl_File_Chooser.H>

#include "ui.h"
#include "version.h"
#include "../shared.src/icons.h"


// global pointer to ui state:
ui_state *g_ui_state  = NULL;

ui_state::ui_state(const scene &S, int argc, char *argv[], std::string resource_dir) :
	m_mouse_mode(MM_NONE), 
	m_mouse_mode_button(NULL),
	m_key_down(-1),

	m_scale_image(NULL),
	m_rotate_image(NULL),
	m_translate_parallel_image(NULL),
	m_translate_orthogonal_image(NULL),

	m_camera_fov_image(NULL),
	m_camera_rotate_image(NULL),
	m_camera_orbit_image(NULL),
	m_camera_translate_parallel_image(NULL),
	m_camera_translate_orthogonal_image(NULL),

	m_spotlight_image(NULL),

	m_select_image(NULL)
	{

	g_ui_state = this;

	m_scene = S;

	m_scale_image = get_icon_image("scale");
	m_rotate_image = get_icon_image("rotate");
	m_translate_parallel_image = get_icon_image("translate_parallel");
	m_translate_orthogonal_image = get_icon_image("translate_orthogonal");

	m_camera_fov_image = get_icon_image("camera_fov");
	m_camera_rotate_image = get_icon_image("camera_rotate");
	m_camera_orbit_image = get_icon_image("camera_orbit");
	m_camera_translate_parallel_image = get_icon_image("camera_translate_parallel");
	m_camera_translate_orthogonal_image = get_icon_image("camera_translate_orthogonal");

	m_spotlight_image = get_icon_image("spotlight");

	m_select_image = get_icon_image("select");	

	// create main window
	static char main_window_name[256];
	sprintf(main_window_name, "Ray Tracer [%s]", g_ray_version);
	m_main_window = new main_window(1024, 768, main_window_name);

	// create menu bar (in seperate file ui_menu.cpp)
	create_menu();

	// create button bar (in seperator group)
	create_button_group();

	// create GL window
	m_model_window = new modeling_window(0, m_buttons_group->y() + m_buttons_group->h(), 
		m_main_window->w(), m_main_window->h() - (m_buttons_group->y() + m_buttons_group->h()));

	// close main window
	m_main_window->resizable(m_model_window);
	m_main_window->end();
	m_main_window->show(argc, argv);

	// create properties window
	m_properties_window = new properties_window(700, 350, "Properties");
	m_properties_window->show(argc, argv);
	m_properties_window->hide();

	// create properties window
	m_render_window = new render_window(400, 300, "Render");
	m_render_window->show(argc, argv);
	m_render_window->hide();

	// set mouse mode to select:
	button_callback(m_buttons[10], MM_SELECT);

	// update various stuff that must be set according to current scene
	new_scene_loaded();
}

/*ui_state::ui_state(const ui_state &ui) {
	throw std::string("please don't copy this class (ui_state)");
}*/

ui_state::~ui_state() {

	if (m_scale_image) delete m_scale_image;
	if (m_rotate_image) delete m_rotate_image;
	if (m_translate_parallel_image) delete m_translate_parallel_image;
	if (m_translate_orthogonal_image) delete m_translate_orthogonal_image;

	if (m_camera_fov_image) delete m_camera_fov_image;
	if (m_camera_rotate_image) delete m_camera_rotate_image;
	if (m_camera_orbit_image) delete m_camera_orbit_image;
	if (m_camera_translate_parallel_image) delete m_camera_translate_parallel_image;
	if (m_camera_translate_orthogonal_image) delete m_camera_translate_orthogonal_image;

	if (m_spotlight_image) delete m_spotlight_image;

	if (m_select_image) delete m_select_image;

	g_ui_state = NULL;
	
}

ui_state &ui_state::operator=(const ui_state &ui) {
	throw std::string("please don't assign this class (ui_state)");	
}

void ui_state::run() {
	Fl::run();
}

void ui_button_callback(Fl_Widget *W, long data) {
	g_ui_state->button_callback(W, data);
}

void ui_state::create_button_group() {
	m_buttons_group = new Fl_Group(0, m_menu_bar->y() + m_menu_bar->h(), m_main_window->w(), 70);
	m_buttons_group->box(FL_ENGRAVED_BOX);
	m_buttons_group->begin();

	int x = 3;
	int y = m_buttons_group->y() + 3;
	int bw = 64;
	int bh = 64;
	
	// a little table to simplify button creation:
	struct {
		Fl_Button **ptr;
		const char *name;
		Fl_Image *image;
		mouse_modes mode;
		int sep_x; // x seperator
	} buttons[] = {
		{&m_buttons[0], "scale (q)", m_scale_image, MM_SCALE, 0},
		{&m_buttons[1], "translate parallel (w)", m_translate_parallel_image, MM_TRANSLATE_PARALLEL, 0},
		{&m_buttons[2], "translate orthogonal (e)", m_translate_orthogonal_image, MM_TRANSLATE_ORTHOGONAL, 00},
		{&m_buttons[3], "rotate (r)", m_rotate_image, MM_ROTATE, 10},

		{&m_buttons[4], "camera field of view / spotlight (a)", m_camera_fov_image, MM_CAMERA_FOV, 0},
		{&m_buttons[5], "translate camera parallel (s)", m_camera_translate_parallel_image, MM_CAMERA_TRANSLATE_PARALLEL, 0},
		{&m_buttons[6], "translate orthogonal (d)", m_camera_translate_orthogonal_image, MM_CAMERA_TRANSLATE_ORTHOGONAL, 0},
		{&m_buttons[7], "rotate camera (f)", m_camera_rotate_image, MM_CAMERA_ROTATE, 0},
		{&m_buttons[8], "orbit camera (g)",m_camera_orbit_image, MM_CAMERA_ORBIT, 10},
		
		{&m_buttons[9], "spotlight (z)", m_spotlight_image, MM_SPOTLIGHT, 10},

		{&m_buttons[10], "select (x)", m_select_image, MM_SELECT, 10},

		{NULL, NULL,  m_rotate_image, MM_ROTATE, 0} // NULLs signal end of array
	};

	// create all buttons:
	int i = 0;
	while (buttons[i].ptr) {
		(*(buttons[i].ptr)) = new Fl_Button(x, y, bw, bh);
		(*(buttons[i].ptr))->callback(ui_button_callback, buttons[i].mode);
		(*(buttons[i].ptr))->tooltip(buttons[i].name);
		(*(buttons[i].ptr))->image(*(buttons[i].image));

		x += (*(buttons[i].ptr))->w() + 2 + buttons[i].sep_x;
		i++;
	}
	
	// create dummy group to prevent resizing of the buttons:
	Fl_Group *dummy = new Fl_Group(x, y, m_buttons_group->w() - x -6, bh);
	dummy->begin();
	dummy->end();


	m_buttons_group->resizable(dummy);
	m_buttons_group->end();
}

void ui_state::button_callback(Fl_Widget *W, long data) {
	switch (data) {
	case MM_ROTATE:
	case MM_TRANSLATE_PARALLEL:
	case MM_TRANSLATE_ORTHOGONAL:
	case MM_SCALE:
	case MM_CAMERA_FOV:
	case MM_CAMERA_ROTATE:
	case MM_CAMERA_ORBIT:
	case MM_CAMERA_TRANSLATE_PARALLEL:
	case MM_CAMERA_TRANSLATE_ORTHOGONAL:
	case MM_SPOTLIGHT:
	case MM_SELECT:
		{
			// reset old button to normal state:
			if (m_mouse_mode_button) {
				m_mouse_mode_button->box(FL_UP_BOX);
				m_mouse_mode_button->redraw();
			}

			if (m_mouse_mode_button == W) {
				// no mouse mode at all
				m_mouse_mode = MM_NONE;
				m_mouse_mode_button = NULL;
			}
			else {
				// set current button to 'down' state
				m_mouse_mode = (mouse_modes)data;
				m_mouse_mode_button = W;
				m_mouse_mode_button->box(FL_DOWN_BOX);
				m_mouse_mode_button->redraw();
			}

			scene_changed();
		}
		break;
	default:
		break;
	}

}

void ui_state::add_model() {
	// get filename from user:
	std::string filename;
	{
		const char *tmp = fl_file_chooser("Select .rtm file", "*.rtm", NULL);
		if (tmp) filename = tmp;
	}

	if (filename.size()) {
		// try to load the model
		try {
			m_scene.add_model(filename);
			scene_changed();
		}
		catch (std::string &S) {
			fl_alert(("Something went wrong while loading " + filename + ":\n" + S).c_str());
		}
		catch (...) {
			fl_alert(("Unknown exception while loading " + filename).c_str());
		}
	}
}

/// Sets the key that is currently being pushed (used for quick 'mode change')
void ui_state::set_key_down(int k) {
	Fl_Button *B;

	// check current state, return if no change . . .
	if (m_key_down == k) return;

	// first undo the current button state:
	B = get_button_for_key(m_key_down);
	if (B) {
		B->color(FL_GRAY);
		B->redraw();
	}

	// set new value
	m_key_down = k;

	// set button state (color it red)
	B = get_button_for_key(m_key_down);
	if (B) {
		B->color(FL_RED);
		B->redraw();
	}


}

/// Resets the key that is currently being pushed to none (0) (used for quick 'mode change')
void ui_state::set_no_key_down() {
	set_key_down(0);
}

/// Returns the key currently being pushed
int ui_state::get_key_down() const {
	return m_key_down;
}

Fl_Button *ui_state::get_button_for_key(int k) {

	/*
	All the tooltips for the buttons end with their shortcut character "(x)", 
	so use that info to find the right button.
	*/
	for (int i = 0; i < m_nb_buttons; i++) {
		// get tooltip, check string length
		const char *tt;
		int l;
		if (m_buttons[i] &&
			(tt = m_buttons[i]->tooltip()) &&
			( (l = (int)strlen(tt)) > 2) ) 
		{
			// OK: now check end of the tooltip :
			if ((tt[l-3] == '(') && (tt[l-2] == k) && (tt[l-1] == ')'))
				return m_buttons[i];
		}
	}

	// button not found:
	return NULL;
}

mouse_modes ui_state::get_mouse_mode() {
	Fl_Button *B = get_button_for_key(get_key_down());
	if (B) {
		return (mouse_modes)B->argument();
	}
	else return m_mouse_mode;
}

void ui_state::main_window_closed() {
	if (m_main_window) {
		delete m_main_window;
		m_main_window = NULL;
	}
	if (m_properties_window) {
		delete m_properties_window;
		m_properties_window = NULL;
	}
	if (m_render_window) {
		delete m_render_window;
		m_render_window = NULL;
	}
}

void ui_state::update_selected_model() {
	try {
		m_properties_window->update_selected_model(get_scene().get_selected_model());
	}
	catch (std::string &) {}
}

void ui_state::update_selected_light_source() {
	try {
		m_properties_window->update_selected_light_source(get_scene().get_selected_light_source());
	}
	catch (std::string &) {}
}

void ui_state::update_selected_none() {
	m_properties_window->update_selected_none();
}

void ui_state::new_scene_loaded() {
	update_current_camera();
	update_display_lights();
}

void ui_state::disable_modeling() {
	m_properties_window_was_visible = m_properties_window->visible() != 0;
	m_properties_window->hide();
	m_main_window->hide();
}

void ui_state::enable_modeling() {
	if (m_properties_window_was_visible)
		m_properties_window->show();
	m_main_window->show();
}







