
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <stdlib.h>
#include <string.h>

#include <string>

// disable some annoying warnings in VC++ due to FLTK (1.1.6)
#pragma warning( disable : 4311 4312 4244 )

#include <FL/Fl_Box.H>

#include "properties_window.h"

#include "../shared.src/fltk_stuff.h"
#include "../shared.src/color.h"
#include "texture.h"
#include "model.h"
#include "ui.h"


namespace {

/// Creates a label at [x,y], size: w x h
Fl_Box *create_label(int x, int y, int w, int h, const char *txt) {
	Fl_Box *label = new Fl_Box(x, y, w, h, txt);
	label->align(FL_ALIGN_LEFT | FL_ALIGN_INSIDE);
	label->box(FL_FLAT_BOX);
	return label;
}

void pw_callback(Fl_Widget *W, void *Data) {
	((properties_window*)Data)->callback(W, Data);
}
}

pw_color_chooser::pw_color_chooser(properties_window *W, bool have_color, bool have_texture, int x, int y, int w, int h, const char *color_name) {
	const int color_chooser_height = 95;
	const int label_height = 25;

	int start_y = y;

	// label
	if (have_color) {
		m_label = create_label(x, y, w, label_height, color_name);
		y += label_height + 5;
	}
	else {
		m_label = create_label(x, y, 0, 0, color_name);
		m_label->hide();
		y += 0;
	}

	m_color_chooser = NULL;
	m_texture_check_button = NULL;
	m_texture_filename_input = NULL;
	m_texture_browse_button = NULL;

	if (have_color) {
		// color chooser
		m_color_chooser = new Fl_Color_Chooser(x, y, w, color_chooser_height);
		m_color_chooser->rgb(0.8, 0.6, 0.4);
		m_color_chooser->callback(pw_callback, W);

		y += color_chooser_height + 5;
	}

	if (have_texture) {
		// check button texture yes / no
		m_texture_check_button = new Fl_Check_Button(x, y, 20, 20);
		m_texture_check_button->callback(pw_callback, W);

		// texture filename input
		m_texture_filename_input = new Fl_Input(
			m_texture_check_button->x() + m_texture_check_button->w() + 5, 
			y, w - 25 - 30, m_texture_check_button->h());
		m_texture_filename_input->callback(pw_callback, W);

		// texture filename browse button
		m_texture_browse_button = new Fl_Button(
			m_texture_filename_input->x() + m_texture_filename_input->w() + 5, 
			y, 25, m_texture_check_button->h(), "...");
		m_texture_browse_button->callback(pw_callback, W);

		m_height = m_texture_check_button->y() + m_texture_check_button->h() + 5 - start_y;
	}
	else {
		m_height = y - start_y;
	}


}


void pw_color_chooser::callback(Fl_Widget *W, void *Data) {
	properties_window *PW = (properties_window *)Data;
	bool update = false; // set to true if an update of UI widgets is required

	if (W == m_color_chooser) {
		PW->set_color(m_label->label(),
			color(
				(float)m_color_chooser->r(), 
				(float)m_color_chooser->g(), 
				(float)m_color_chooser->b()));
	}
	else if (W == m_texture_check_button) {
		PW->set_texture_enabled(m_label->label(), m_texture_check_button->value() != 0);
	}
	else if (W == m_texture_browse_button) {
		// todo: this code causes trouble in debug mode, FLTK, windows, microsoft compiler . . .
		const char *filename = fl_file_chooser_no_preview("Select png image", "*.png", NULL, 0);
		if (filename) {
			PW->set_texture(m_label->label(), filename);
			update = true;
		}
	}
	else if (W == m_texture_filename_input) {
		PW->set_texture(m_label->label(), m_texture_filename_input->value());
		update = true;
	}

	// update the widgets if required:
	if (update) {
		try {
			PW->update_selected_model(g_ui_state->get_scene().get_selected_model());
		} catch (const std::string &) {};
		try {
			PW->update_selected_light_source(g_ui_state->get_scene().get_selected_light_source());
		} catch (const std::string &) {};
	}
}

void pw_color_chooser::update(const color &c, const texture &t) {
	if (m_color_chooser)
		m_color_chooser->rgb(c.r(), c.g(), c.b());
	m_texture_check_button->value(t.enabled());
	if (t.valid())
		m_texture_check_button->activate();
	else m_texture_check_button->deactivate();

	const int l = 1023;
	static char texture_filename[l+1];
	if (t.get_filename().size() < l)
		strcpy(texture_filename, t.get_filename().c_str());
	else memcpy(texture_filename, t.get_filename().c_str(), l), texture_filename[0] = 0;

	m_texture_filename_input->value(texture_filename);
}

void pw_color_chooser::update(const color &c) {
	if (m_color_chooser)
		m_color_chooser->rgb(c.r(), c.g(), c.b());
}

properties_window::properties_window(int w, int h, const char* label/*=0*/) : Fl_Window(w, h, label) {
	setup_widgets();
}

properties_window::properties_window(int x, int y, int w, int h, const char* label/*=0*/) : Fl_Window(x, y, w, h, label) {
	setup_widgets();
}

properties_window::~properties_window() {
	if (m_model_ambient_color_chooser)
		delete m_model_ambient_color_chooser;
	if (m_model_diffuse_color_chooser)
		delete m_model_diffuse_color_chooser;
	if (m_model_specular_color_chooser)
		delete m_model_specular_color_chooser;

	if (m_shininess_color_chooser)
		delete m_shininess_color_chooser;
	if (m_bump_map_color_chooser)
		delete m_bump_map_color_chooser;

	if (m_light_source_ambient_color_chooser)
		delete m_light_source_ambient_color_chooser;
	if (m_light_source_diffuse_color_chooser)
		delete m_light_source_diffuse_color_chooser;
	if (m_light_source_specular_color_chooser)
		delete m_light_source_specular_color_chooser;
}


void properties_window::setup_widgets() {
	const int color_chooser_width = 200;

	// Set callback of the properties window itself.
	// This callback will be called when the window manager wants to close this window.
	// In response pw_callback() will call hide();
	Fl_Window::callback(pw_callback, this);

	/*
	create the model group:
	*/
	m_model_group = new Fl_Group(0, 0, w(), h());

	int x = 5;
	int y = 5;

	{
		// ambient color
		m_model_ambient_color_chooser = new pw_color_chooser(this, true, true, x, y, color_chooser_width, 100, "Ambient:");
		x += color_chooser_width + 25;

		// diffuse color
		m_model_diffuse_color_chooser = new pw_color_chooser(this, true, true, x, y, color_chooser_width, 100, "Diffuse:");
		x += color_chooser_width + 25;

		// specular color
		m_model_specular_color_chooser = new pw_color_chooser(this, true, true, x, y, color_chooser_width, 100, "Specular:");
		x += color_chooser_width + 25;

		y += m_model_ambient_color_chooser->h() + 25;

		x = 5;


		// shininess factor:
		Fl_Box *_B = create_label(x, y, 150, 20, "shininess: ");
		m_shininess_slider = new Fl_Value_Slider(x + _B->w(), y, w() - 10 - _B->w() - color_chooser_width - 5, 20);
		m_shininess_slider->type(FL_HOR_SLIDER);
		m_shininess_slider->range(0.0, 250.0);
		m_shininess_slider->when(FL_WHEN_CHANGED | FL_WHEN_RELEASE);
		m_shininess_slider->callback(pw_callback, this);

		// shininess texture
		m_shininess_color_chooser = new pw_color_chooser(this, false, true, m_shininess_slider->x() + m_shininess_slider->w() + 5, y, color_chooser_width, 20, "Shininess:");

		y += m_shininess_slider->h() + 5;

		create_label(x, y, 150, 20, "bumpyness: ");
		m_bumpyness_slider = new Fl_Value_Slider(x + _B->w(), y, w() - 10 - _B->w() - color_chooser_width - 5, 20);
		m_bumpyness_slider->type(FL_HOR_SLIDER);
		m_bumpyness_slider->range(0.0, 10.0);
		m_bumpyness_slider->when(FL_WHEN_CHANGED | FL_WHEN_RELEASE);
		m_bumpyness_slider->callback(pw_callback, this);

		// bump map texture
		m_bump_map_color_chooser = new pw_color_chooser(this, false, true, m_bumpyness_slider->x() + m_bumpyness_slider->w() + 5, y, color_chooser_width, 20, "Bumpmap:");

		y += m_bumpyness_slider->h() + 5;

		// direct lighting factor:
		create_label(x, y, 150, 20, "direct lighting factor: ");
		m_direct_lighting_slider = new Fl_Value_Slider(m_shininess_slider->x(), y, m_shininess_slider->w(), m_shininess_slider->h());
		m_direct_lighting_slider->type(FL_HOR_SLIDER);
		m_direct_lighting_slider->range(0.0, 1.0);
		m_direct_lighting_slider->when(FL_WHEN_CHANGED | FL_WHEN_RELEASE);
		m_direct_lighting_slider->callback(pw_callback, this);

		y += m_shininess_slider->h() + 5;

		// reflection factor:
		create_label(x, y, 150, 20, "reflection factor: ");
		m_reflection_slider = new Fl_Value_Slider(m_shininess_slider->x(), y, m_shininess_slider->w(), m_shininess_slider->h());
		m_reflection_slider->type(FL_HOR_SLIDER);
		m_reflection_slider->range(0.0, 1.0);
		m_reflection_slider->when(FL_WHEN_CHANGED | FL_WHEN_RELEASE);
		m_reflection_slider->callback(pw_callback, this);

		y += m_shininess_slider->h() + 5;

		// refraction factor:
		create_label(x, y, 150, 20, "refraction factor: ");
		m_refraction_slider = new Fl_Value_Slider(m_shininess_slider->x(), y, m_shininess_slider->w(), m_shininess_slider->h());
		m_refraction_slider->type(FL_HOR_SLIDER);
		m_refraction_slider->range(0.0, 1.0);
		m_refraction_slider->when(FL_WHEN_CHANGED | FL_WHEN_RELEASE);
		m_refraction_slider->callback(pw_callback, this);

		y += m_shininess_slider->h() + 5;

		// refractive index:
		create_label(x, y, 150, 20, "reflective index: ");
		m_refractive_index_slider = new Fl_Value_Slider(m_shininess_slider->x(), y, m_shininess_slider->w(), m_shininess_slider->h());
		m_refractive_index_slider->type(FL_HOR_SLIDER);
		m_refractive_index_slider->range(0.0, 10.0); // what's that max?
		m_refractive_index_slider->when(FL_WHEN_CHANGED | FL_WHEN_RELEASE);
		m_refractive_index_slider->callback(pw_callback, this);

		y += m_shininess_slider->h();
	}
	m_model_group->end();
	/*
	model group created
	*/


	/*
	create the light source group:
	*/
	m_light_source_group = new Fl_Group(0, 0, w(), h());

	x = 5;
	y = 5;

	{
		// ambient color
		m_light_source_ambient_color_chooser = new pw_color_chooser(this, true, false, x, y, color_chooser_width, 100, "Ambient:");
		x += color_chooser_width + 25;

		// diffuse color
		m_light_source_diffuse_color_chooser = new pw_color_chooser(this, true, false, x, y, color_chooser_width, 100, "Diffuse:");
		x += color_chooser_width + 25;

		// specular color
		m_light_source_specular_color_chooser = new pw_color_chooser(this, true, false, x, y, color_chooser_width, 100, "Specular:");
		x += color_chooser_width + 25;

		y += m_light_source_ambient_color_chooser->h() + 25;

		x = 5;

		// spot exponent:
		Fl_Box *_B = create_label(x, y, 150, 20, "spot exponent: ");
		m_spot_exponent_slider = new Fl_Value_Slider(x + _B->w(), y, w() - 10 - _B->w(), 20);
		m_spot_exponent_slider->type(FL_HOR_SLIDER);
		m_spot_exponent_slider->range(0.0, 250.0);
		m_spot_exponent_slider->when(FL_WHEN_CHANGED | FL_WHEN_RELEASE);
		m_spot_exponent_slider->callback(pw_callback, this);
	}

	m_light_source_group->end();
	/*
	light source group created
	*/

	// select none:
	update_selected_none();

}


void properties_window::callback(Fl_Widget *W, void *Data) {
	if (W == this) {
		hide();
		g_ui_state->set_properties_window_shown(false);
	}
	else if (W == m_shininess_slider) {
		g_ui_state->get_scene().get_selected_model().set_shininess(m_shininess_slider->value());
		g_ui_state->scene_changed();
	}
	else if (W == m_bumpyness_slider) {
		g_ui_state->get_scene().get_selected_model().set_bumpyness(m_bumpyness_slider->value());
		g_ui_state->scene_changed();
	}
	else if (W == m_direct_lighting_slider) {
		g_ui_state->get_scene().get_selected_model().set_direct_lighting_factor(m_direct_lighting_slider->value());
		g_ui_state->scene_changed();
	}
	else if (W == m_reflection_slider) {
		g_ui_state->get_scene().get_selected_model().set_reflection_factor(m_reflection_slider->value());
		g_ui_state->scene_changed();
	}
	else if (W == m_refraction_slider) {
		g_ui_state->get_scene().get_selected_model().set_refraction_factor(m_refraction_slider->value());
		g_ui_state->scene_changed();
	}
	else if (W == m_refractive_index_slider) {
		g_ui_state->get_scene().get_selected_model().set_refractive_index(m_refractive_index_slider->value());
		g_ui_state->scene_changed();
	}
	else if (W == m_spot_exponent_slider) {
		g_ui_state->get_scene().get_selected_light_source().set_spot_exponent(m_spot_exponent_slider->value());
		g_ui_state->scene_changed();
	}
	else {
		m_model_ambient_color_chooser->callback(W, Data);
		m_model_diffuse_color_chooser->callback(W, Data);
		m_model_specular_color_chooser->callback(W, Data);

		m_shininess_color_chooser->callback(W, Data);
		m_bump_map_color_chooser->callback(W, Data);

		m_light_source_ambient_color_chooser->callback(W, Data);
		m_light_source_diffuse_color_chooser->callback(W, Data);
		m_light_source_specular_color_chooser->callback(W, Data);
	}
}

void properties_window::update_selected_model(const model &m) {
	// label
	set_label("properties: model " + m.get_mesh()->get_filename());

	// update colors / textures
	m_model_ambient_color_chooser->update(m.get_ambient_color(), m.get_ambient_texture());	
	m_model_diffuse_color_chooser->update(m.get_diffuse_color(), m.get_diffuse_texture());	
	m_model_specular_color_chooser->update(m.get_specular_color(), m.get_specular_texture());	

	// update bumpmap & shininess
	m_shininess_color_chooser->update(::color(), m.get_shininess_texture());	
	m_bump_map_color_chooser->update(::color(), m.get_bump_map_texture());	

	// update various factors / indices
	m_shininess_slider->value(m.get_shininess());
	m_bumpyness_slider->value(m.get_bumpyness());
	m_direct_lighting_slider->value(m.get_direct_lighting_factor());
	m_reflection_slider->value(m.get_reflection_factor());
	m_refraction_slider->value(m.get_refraction_factor());
	m_refractive_index_slider->value(m.get_refractive_index());

	// show / hide the right group:
	m_model_group->show();
	m_light_source_group->hide();
}
	
void properties_window::update_selected_light_source(const light_source &l) {
	// label
	set_label("properties: light");

	// update colors / textures
	m_light_source_ambient_color_chooser->update(l.get_ambient_color());	
	m_light_source_diffuse_color_chooser->update(l.get_diffuse_color());	
	m_light_source_specular_color_chooser->update(l.get_specular_color());	

	m_spot_exponent_slider->value(l.get_spot_exponent());

	// show / hide the right group:
	m_model_group->hide();
	m_light_source_group->show();
}

void properties_window::update_selected_none() {
	// label
	set_label("properties: no object selected");

	// show / hide the right groups:
	m_model_group->hide();
	m_light_source_group->hide();
}


void properties_window::set_color(const std::string &what, const ::color &c) {
	// try to get model / light source & set the color
	try {
		g_ui_state->get_scene().get_selected_model().set_color(what, c);
	}
	catch (std::string &) {
		try {
			g_ui_state->get_scene().get_selected_light_source().set_color(what, c);
		}
		catch (std::string &) {}
	}
	g_ui_state->scene_changed();
}

bool properties_window::set_texture(const std::string &what, const char *filename) {
	bool s = false;
	// try to get model / light source & set the color
	try {
		s = g_ui_state->get_scene().get_selected_model().set_texture(what, filename);
	}
	catch (std::string &) {
	}
	g_ui_state->scene_changed();
	return s;
}

void properties_window::set_texture_enabled(const std::string &what, bool e) {
	// try to get model / light source & set the color
	try {
		g_ui_state->get_scene().get_selected_model().set_texture_enabled(what, e);
	}
	catch (std::string &) {
	}
	g_ui_state->scene_changed();
}


void properties_window::set_label(const std::string &pw_label_str) {
	const int max_pw_label = 1023;
	static char pw_label[max_pw_label + 1];
	if (pw_label_str.size() > max_pw_label) {
		memcpy(pw_label, pw_label_str.c_str(), max_pw_label);
		pw_label[max_pw_label] = 0;
	}
	else strcpy(pw_label, pw_label_str.c_str());
	label(pw_label);
}


