
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <stdlib.h>
#include <string.h>

// disable some annoying warnings in VC++ due to FLTK (1.1.6)
#pragma warning( disable : 4311 4312 4244 )

#include <FL/Fl.H>
#include <FL/Fl_File_Chooser.H>
#include <FL/fl_ask.H>
#include "render_window.h"
#include "ui.h"
#include "../shared.src/pngfe.h"

namespace {

/// Creates a label at [x,y], size: w x h
Fl_Box *create_label(int x, int y, int w, int h, const char *txt) {
	Fl_Box *label = new Fl_Box(x, y, w, h, txt);
	label->align(FL_ALIGN_LEFT | FL_ALIGN_INSIDE);
	label->box(FL_FLAT_BOX);
	return label;
}

void rw_callback(Fl_Widget *W, void *Data) {
	((render_window*)Data)->callback(W, Data);
}

void rw_idle_callback(void *Data) {
	((render_window*)Data)->timer_callback();
}
}

render_window::render_window(int w, int h, const char* label/*=0*/) : Fl_Double_Window(w, h, label) {
	setup_widgets();
}

render_window::render_window(int x, int y, int w, int h, const char* label/*=0*/) : Fl_Double_Window(x, y, w, h, label) {
	setup_widgets();
}

render_window::~render_window() {
}


void render_window::setup_widgets() {
	m_image = NULL;
	m_render_buffer_width = m_render_buffer_height = -1;
	m_render_scanline = m_scanline_step = -1;
	m_nb_scanline_rendered = 0;

	// Set callback of the properties window itself.
	// This callback will be called when the window manager wants to close this window.
	// In response rw_callback() will call hide() and stop the render (if any)
	Fl_Double_Window::callback(rw_callback, this);


	int x = 5, y = 5;
	int iw = (w() - 10) /3;

	// width:
	Fl_Box *LW = create_label(x, y, 50, 20, "width:");
	m_width_input = new Fl_Int_Input(x + LW->w(), y, LW->w(),LW->h());
	m_width_input->value("128");
	x += m_width_input->w() + m_width_input->x() - LW->x() + 10;

	// height:
	Fl_Box *LH = create_label(x, y, 50, 20, "height:");
	m_height_input = new Fl_Int_Input(x + LH->w(), y, LH->w(), LH->h());
	m_height_input->value("128");
	x += m_height_input->w() + m_height_input->x() - LH->x() + 10;


	// multisample
	m_multisample_slider = new Fl_Value_Slider(x , y, w() - x - 5, 20, "multisample");
	m_multisample_slider->type(FL_HOR_SLIDER);
	m_multisample_slider->range(1.0, 8.0);
	m_multisample_slider->step(1.0);
	m_multisample_slider->value(1.0);
	m_multisample_slider->when(FL_WHEN_CHANGED | FL_WHEN_RELEASE);

	x = 5;
	y += LW->h() + 5;

	// start render:
	iw = w() / 6;
	m_render_button = new Fl_Button(x, y, iw, 20, "render");
	m_render_button->callback(rw_callback, this);
	x += m_render_button->w() + 5;

	// stop render
	m_stop_render_button = new Fl_Button(x, y, iw, 20, "stop");
	m_stop_render_button->callback(rw_callback, this);
	x += m_stop_render_button->w() + 5;

	// save render
	m_save_render_button = new Fl_Button(x, y, iw, 20, "save");
	m_save_render_button->callback(rw_callback, this);
	x += m_stop_render_button->w() + 5;

	x = 5;
	y += m_render_button->h() + 5;

	// render output
	m_render_output_box = new Fl_Box(x, y, w() - x - 5, h() - y - 5);
	m_render_output_box->box(FL_ENGRAVED_BOX);

	// this allows us to resize the window, but disallows the user to so:
	resizable(m_render_output_box);
	size_range(w(), h(), w(), h());
}


void render_window::callback(Fl_Widget *W, void *Data) {
	if (W == this) {
		m_render_output_box->label("Render aborted");
		m_render_output_box->redraw();
		stop_render(); // stop render when window is hidden
		hide();
	}
	else if (W == m_render_button) {
		// start render
		start_render();
	}
	else if (W == m_stop_render_button) {
		// stop render
		m_render_output_box->label("Render aborted");
		m_render_output_box->redraw();
		stop_render();
	}
	else if (W == m_save_render_button) {
		// save if image is available
		char *filename;

		if ((m_render_buffer_width <= 0) || (m_render_buffer_height <= 0))
			fl_alert("No image to save");
		else if (filename = fl_file_chooser("Save image", "*.png", NULL)) {
			try {
				save_png(filename, &(m_render_buffer[0]), m_render_buffer_width, m_render_buffer_height);
			} catch (const std::string &S) {
				fl_alert((std::string("Could not save ") + filename + ":\n" + S).c_str());
			}
		}
	}


}

void  render_window::timer_callback() {
	if (rendering()) {

		scene &S = g_ui_state->get_scene();

//		printf("Render scanline %d\n", m_render_scanline);

		static char buf[256];
		sprintf(buf, "Rendering scanline %d (%2.1f%% done)", 
			m_render_scanline, 
			100.0f * (float)m_nb_scanline_rendered / (float)m_render_buffer_height);
		m_render_output_box->label(buf);
		m_render_output_box->redraw();

		// render scanline
		S.render(S.get_current_camera_idx(), &(m_render_buffer[0]), 
			m_render_buffer_width, m_render_buffer_height, 
			(int)m_multisample_slider->value(), 0, // 0 = verbosity
			NULL, 4, // NULL = no intersection info, 4 = max recursion
			m_render_scanline);
		m_scanline_rendered[m_render_scanline] = true;
		m_nb_scanline_rendered++;

		// copy scanline to all scanlines below it for quick preview
		for (int i = 1; 
			(i < m_scanline_step) && ((m_render_scanline + i)<m_render_buffer_height); 
			i++) {

			memcpy(
				&(m_render_buffer[0]) + m_render_buffer_width * 3 * (m_render_scanline + i),
				&(m_render_buffer[0]) + m_render_buffer_width * 3 * m_render_scanline,
				m_render_buffer_width * 3);
		}

		// move to the next scanline
		while (rendering() && m_scanline_rendered[m_render_scanline]) {
			do {
				m_render_scanline += m_scanline_step;
			} while ((m_render_scanline < m_render_buffer_height) &&
				m_scanline_rendered[m_render_scanline]);

			// see if we are done (entirely, or at this 'step level)
			if (m_render_scanline >= m_render_buffer_height) {
				if (m_scanline_step == 1) { // done entirely
					sprintf(buf, "Render done");
					m_render_output_box->label(buf);
					stop_render();
				}
				else { // move to next step level
					m_scanline_step >>= 1;
					m_render_scanline = 0;
				}
			}
		}

		/*// stop render when all done, or switch to odd scanlines:
		if (m_render_scanline >= m_render_buffer_height) {
			if (m_render_scanline & 1) {
				sprintf(buf, "Render done");
				m_render_output_box->label(buf);
				stop_render();
			}
			else m_render_scanline = 1;
		}*/

		// update image
		Fl_Image *prev_image = m_image;
		m_image = new Fl_RGB_Image(&(m_render_buffer[0]), m_render_buffer_width, m_render_buffer_height);
		m_render_output_box->image(m_image);
		m_render_output_box->redraw();

		if (prev_image) delete prev_image;
	}


	if (!rendering()) {
		// auto switch off:
		Fl::remove_idle(rw_idle_callback, this);
	}
}

bool render_window::rendering() const {
	return (m_render_scanline >= 0);	
}  

void render_window::start_render() {
	if (rendering()) return;

	int width = atoi(m_width_input->value());
	int height = atoi(m_height_input->value());

	if ((width > 0) && (height > 0)) {
		// resize render buffer, set it to gray
		m_render_buffer.resize(width * height * 3);
		memset(&(m_render_buffer[0]), 128, width * height * 3);
		m_render_buffer_width = width;
		m_render_buffer_height = height;	

		// init stuff for 'incremental' rendering
		m_render_scanline = 0;
		for (m_scanline_step = 1; m_scanline_step < m_render_buffer_height; m_scanline_step = m_scanline_step << 1);
		m_scanline_rendered.resize(m_render_buffer_height);
		for (int i = 0; i < m_render_buffer_height; i++) m_scanline_rendered[i] = false;
		m_nb_scanline_rendered = 0;

		// resize so image fits in window, but disallows the user to so:
		int ww = (width < 300) ? 300 : width;
		int wh = ((height +20) < 300) ? 300 : (height+20);
		int newW = w() + ww - (m_render_output_box->w() - 10);
		int newH = h() + wh - (m_render_output_box->h() - 10);
		size_range(newW, newH, newW, newH); // this forces 
		resize(x(), y(), newW, newH);
		redraw(); // avoid redraw errors due to resizing . . .

		// when idle, call this function;
		// it is used to render a scanline at a time
		Fl::add_idle(rw_idle_callback, this);

		m_render_button->deactivate();
		m_width_input->deactivate();
		m_height_input->deactivate();
		m_multisample_slider->deactivate();

		g_ui_state->disable_modeling();
	}

	scene &S = g_ui_state->get_scene();
	S.render_init();

	
	c3ga::g2Profiling::reset();
}

void render_window::stop_render() {
	if (!rendering()) return;
	
	c3ga::g2Profiling::save();

	m_render_scanline = -1;

	m_render_button->activate();
	m_width_input->activate();
	m_height_input->activate();
	m_multisample_slider->activate();

	g_ui_state->enable_modeling();
	show();
}

