
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef _SCENE_H_
#define _SCENE_H_

#include <vector>
#include <map>
#include <string>

#include "classes.h"
#include "model.h"
#include "light_source.h"
#include "camera.h"
#include "mesh.h"

class scene {
	friend class light_source; // it calls find_intersection() and get_light_idx()
public:
	scene(); ///< null constructor
	scene(const scene &s); ///< copy constructor
	scene(const std::string &filename); ///< load file constructor
	~scene(); ///< desctructor

	scene &operator=(const scene &s); ///< copy operator

	/// returns the number of models in the scene
	inline int get_nb_models() const {return (int)m_model.size();}
	/// returns the model at index \param idx . Array bounds not checked
	inline const model& get_model(int idx) const {return m_model[idx];}
	inline model& get_model(int idx) {return m_model[idx];}
	/// adds model loaded from \param filename to this scene
	void add_model(const std::string &filename);
	/// adds \param M
	void add_model(const model &M);
	/// duplicates a model \param M
	void duplicate_model(const model &M);

	const camera &get_camera(int idx) const; ///< returns m_camera[idx]; throws std::string when index out of range (const)
	camera &get_camera(int idx); ///< returns m_camera[idx]; throws std::string when index out of range

	void set_camera(int idx, const camera &C); ///< sets m_camera[idx] to C; throws std::string when index out of range

	/// returns the number of lights in the scene
	inline int get_nb_light_sources() const {return (int)m_light_source.size();}
	/// returns the light at index \param idx . Array bounds not checked
	inline const light_source& get_light_source(int idx) const {return m_light_source[idx];}
	inline light_source& get_light_source(int idx) {return m_light_source[idx];}
	/// adds a light source to the scene
	void add_light_source(const light_source &L);

	/// Returns the currently active camera (const) ('model' or 'render')
	const camera &get_current_camera() const;
	/// Returns the currently active camera ('model' or 'render')
	camera &get_current_camera();
	
	/// Returns m_current_camera
	int get_current_camera_idx() const;

	/// Sets the currently active camera ('model'(0) or 'render'(1)); throws exception if index out of range
	void set_current_camera(int idx);

	/// Sets selection to none
	void reset_selected_object();

	/// Sets the selected object to \param idx
	void set_selected_object(int idx);

	/// Returns the selected object (or throws exception when no object selected)
	locator &get_selected_object();

	/// Returns the selected light source (or throws exception when no light source selected)
	light_source &get_selected_light_source();

	/// Returns the selected model (or throws exception when no model selected)
	model &get_selected_model();

	/// Removes selected object (does not throw if none selected)
	void remove_selected_object();

	/// Turns this object into a string (e.g., for storing a scene to file)
	std::string to_string(const std::string &basepath) const;

	/**
	Loads the scene in \param filename into this scene
	*/
	void load_rsn(const std::string &filename);

	/// precomputes positions of light, models, camera, so we don't have to do it during rendering
	void render_init();

	/**
	Renders this scene to the provided buffer \param buf .
	The buffer is assume to be \param width by \param height by 3(rgb)
	\param multisample specifies how many sample to take per pixel.
	\param verbosity specifies how much the renderer should print out
	(using printf()) about its progress.
	\param II optionally specifies precomputed intersection info.
	\param max_recursion specifies maximum number of times we 
	should recurse in trace()

	If \param scanline is in range [0, image_height), then only that scanline
	is rendered.

	Calls trace() to do the actual rendering.
	*/
	void render(int camera_idx, unsigned char *image_buffer, 
		int image_width, int image_height, int multisample, int verbosity, 
		const intersection_info *II, int max_recursion, int scanline = -1);

	/// returns m_display_lights
	bool display_lights() const;
	/// sets m_display_lights to \param l
	void set_display_lights(bool l);

	inline c3ga::mv::Float get_refractive_index() const {return 1.0f;}

private:
	/// lights
	std::vector<light_source> m_light_source;
	bool m_display_lights; // true if lights should be displayed in preview mode

	/// cameras
	int m_current_camera; ///< index into the m_camera vector
	std::vector<camera> m_camera;
	/// models
	std::vector<model> m_model;
	// action!

	/**
	Returns model for .rtm file \param filename .
	Checks if rtm \param filename is already in the cache, 
	if so, returns that model. Otherwise calls load_rtm() 
	and updates the cache.

	Throws exception (std::string) if something is wrong
	with the .rtm file.
	*/
	const model &get_rtm(const std::string &filename, const c3ga::EXForm &XF, const c3ga::EXForm &XFi);

	/**
	Loads an rtm file \param filename, initializes it, and stores it in the cache.
	Throws exception (std::string) if something is wrong
	with the .rtm file.
	*/ 
	void load_rtm(const std::string &filename, const c3ga::EXForm &XF, const c3ga::EXForm &XFi);

	/**
	To save some storage & computation time, we load each .rtm file
	only once. The second time a identical .rtm is required, we copy
	it from this cache. 
	
	You access this map by filename of the .rtm file
	*/
	std::map<std::string, model> m_rtm_model_cache;
	std::map<std::string, mesh> m_rtm_mesh_cache;

	/**
	After a copy, all pointers to meshes have to be
	'transfered' from the old object to the new object.
	This function does that (called by copy constructor 
	and assignment operator).
	*/
	void transfer_mesh_pointers();


	/**
	Random seed as remembered by set_random_seed()
	Initially set to 0
	*/
	unsigned long m_random_seed;
	/**
	Sets the random seed for the mersenne twister to \param seed, 
	and remembers \param seed in m_random_seed for future calls 
	to reset_random_seed();
	*/
	void set_random_seed(unsigned long seed);
	/// Sets random seed for mersenne twister to m_random_seed
	void reset_random_seed() const;


	/**
	Traces a ray \param R through this scene. Returns
	the color received from light sources by the ray.
	\param max_recursion is the maximum number of
	recursions allowed. \param II is optional (can be NULL)
	precomputed intersection info.
	*/
	color trace(const ray &R, int max_recursion, const intersection_info *II) const; // why not const?

	/**
	Shade the surface point \param spt a seen from ray \param R .
	Lighting information comes from this scene. Surface information
	from \param spt. The result is the color
	*/
	color shade(const ray &R, const intersection_info *II, const surface_point &spt) const;

	/**
	Reflects ray \param R in the surface specified by \param spt .
	Then calls trace() to determine the color as seen from the ray.
	*/
	color reflect(const ray &R, int max_recursion, const intersection_info *II, const surface_point &spt) const;

	/**
	Reflects ray \param R in the surface specified by \param spt .
	Then calls trace() to determine the color as seen from the ray.
	*/
	color refract(const ray &R, int max_recursion, const intersection_info *II, const surface_point &spt) const;

	/**
	Finds the intersection of ray \param R and any model in the scene.
	Depending on the value of \param find_closest , either a search
	for the closes intersection will be done, or for _any_ intersection.
	For shadow rays, for instance, we only need to know that they intersect
	a model, now what model or where.
	When an intersection is found, and \param find_closest is true, 
	information about that point is stored in \param spt
	*/
	bool find_intersection(const ray &R, surface_point &spt, bool find_closest) const;

	/// returns the index of a light in the m_light_source vector, given a pointer
	inline unsigned int get_light_index(const light_source *L) const {
		return (unsigned int)(L - &(m_light_source[0]));
	}
};

#endif /* _SCENE_H_ */
