
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include "face.h"
#include "mesh.h"
#include "c3ga_util.h"
#include "surface_point.h"
#include "texture.h"
#include "vertex.h"
#include "../shared.src/mt19937-2.h"
#include "../shared.src/constants.h"

face::face() : m_u2(0.0f), m_v2(0.0f), m_w2(0.0f), 
					 m_u_v(0.0f), m_v_w(0.0f), m_w_u(0.0f), 
					 m_u_w2(0.0f) 
{
	for (int i = 0; i < 3; i++)
		m_vtx_idx[i] = 0;

	m_uv_dir[0][0] = m_uv_dir[0][1] = 
		m_uv_dir[1][0] = m_uv_dir[1][1] = 0.0; 
}

face::face(const int vtx_idx[3]) : m_u2(0.0f), m_v2(0.0f), m_w2(0.0f), 
												m_u_v(0.0f), m_v_w(0.0f), m_w_u(0.0f),
												m_u_w2(0.0f) 
{
	for (int i = 0; i < 3; i++)
		m_vtx_idx[i] = vtx_idx[i];

	m_uv_dir[0][0] = m_uv_dir[0][1] = 
		m_uv_dir[1][0] = m_uv_dir[1][1] = 0.0; 
}

face::face(const face &f) : m_u2(f.m_u2), m_v2(f.m_v2), m_w2(f.m_w2), 
										m_u_v(f.m_u_v), m_v_w(f.m_v_w), m_w_u(f.m_w_u), 
										m_u_w2(f.m_u_w2) , m_pl(f.m_pl), m_d(f.m_d), m_valid(f.m_valid)
{
	for (int i = 0; i < 3; i++) {
			m_vtx_idx[i] = f.m_vtx_idx[i];
			m_edgePl[i] = f.m_edgePl[i];
			m_edgeDir[i] = f.m_edgeDir[i];
			m_edgeD[i] = f.m_edgeD[i];
	}

	m_uv_dir[0][0] = f.m_uv_dir[0][0];
	m_uv_dir[0][1] = f.m_uv_dir[0][1];
	m_uv_dir[1][0] = f.m_uv_dir[1][0];
	m_uv_dir[1][1] = f.m_uv_dir[1][1];
}


face::~face() {

}


face &face::operator=(const face &f) {
	if (this != &f) {
		for (int i = 0; i < 3; i++) {
			m_vtx_idx[i] = f.m_vtx_idx[i];
			m_edgePl[i] = f.m_edgePl[i];
			m_edgeDir[i] = f.m_edgeDir[i];
			m_edgeD[i] = f.m_edgeD[i];
		}

		m_u2 = f.m_u2, 
		m_v2 = f.m_v2;
		m_w2 = f.m_w2; 
									
		m_u_v = f.m_u_v; 
		m_v_w = f.m_v_w; 
		m_w_u = f.m_w_u; 
									
		m_u_w2 = f.m_u_w2;

		m_valid = f.m_valid;

		m_uv_dir[0][0] = f.m_uv_dir[0][0];
		m_uv_dir[0][1] = f.m_uv_dir[0][1];
		m_uv_dir[1][0] = f.m_uv_dir[1][0];
		m_uv_dir[1][1] = f.m_uv_dir[1][1];

		m_pl = f.m_pl;
		m_d = f.m_d;
	}
	return *this;
}


void face::precompute_constants(const mesh &m) {
	int i;

	// construct the plane
	m_pl = _bivector(-(get_vtx(1, m).get_pt() - get_vtx(0, m).get_pt()) ^
		(get_vtx(2, m).get_pt() - get_vtx(0, m).get_pt()));

	// Check surface area and normalize plane.
	// If the surface area is 0, then 'm_valid' is set to 0.
	m_valid = (_Float(norm_e2(m_pl)) > 0.0);

	if (m_valid) {
		// normalize plane; compute distance to origin
		m_pl = unit_e(m_pl);
		m_d = _Float(get_vtx(0, m).get_pt() % -dual(m_pl));

		// construct edge planes
		for (i = 0; i < 3; i++) {
			m_edgeDir[i] = get_vtx((i + 1) % 3, m).get_pt() - get_vtx(i % 3, m).get_pt();

			m_edgePl[i] = _bivector(unit_e(dual(m_pl) ^ m_edgeDir[i]));
			m_edgeD[i] = _Float(get_vtx(i % 3, m).get_pt() % -dual(m_edgePl[i]));
		}

		// compute interpolation constants
		m_u2 = _Float(m_edgeDir[0] % m_edgeDir[0]);
		m_v2 = _Float(m_edgeDir[1] % m_edgeDir[1]);
		m_w2 = _Float(m_edgeDir[2] % m_edgeDir[2]);

		m_u_v = _Float(m_edgeDir[0] % m_edgeDir[1]);
		m_v_w = _Float(m_edgeDir[1] % m_edgeDir[2]);
		m_w_u = _Float(m_edgeDir[2] % m_edgeDir[0]);

		m_u_w2 = _Float((m_edgeDir[2] ^ m_edgeDir[0]) % (m_edgeDir[2] ^ m_edgeDir[0])); 
	}
}

bool face::uses_vtx_idx(int vtx_idx) const {
	return ((vtx_idx == get_vtx_idx(0)) || 
		(vtx_idx == get_vtx_idx(1)) || 
		(vtx_idx == get_vtx_idx(2)));
}

void face::compute_local_texture_basis(const mesh &m) {
	e3ga::mv::Float ab[2], ac[2], factor, div;
	int i, j;

	// Todo: could add check for faces which uses texture in a '1d' way and treat that case properly...
	// That is where check for (ab == f * ac) for some f

	// compute how much the texture coordinates change in the 'ab' and 'ac' directions
	ab[0] = get_vtx(1, m).get_tex_uv()[0] - get_vtx(0, m).get_tex_uv()[0];
	ab[1] = get_vtx(1, m).get_tex_uv()[1] - get_vtx(0, m).get_tex_uv()[1];
	ac[0] = get_vtx(2, m).get_tex_uv()[0] - get_vtx(0, m).get_tex_uv()[0];
	ac[1] = get_vtx(2, m).get_tex_uv()[1] - get_vtx(0, m).get_tex_uv()[1];

	// this code inverts a 2x2 matrix from 'ab' and 'ac' and stores the result in 'm_uvDir'
	for (i = 0; i < 2; i++) {
		j = i ^ 1;
		m_uv_dir[i][0] = m_uv_dir[i][1] = 0.0;
		if (ac[j] == 0.0) {
			if (ac[i] == 0.0) continue;
			m_uv_dir[i][1] = 1.0 / ac[i];
		}
		else {
			factor = -ab[j] / ac[j];
			div = ab[i] + ac[i] * factor;
			if (div == 0.0) continue;
			m_uv_dir[i][0] = 1 / div;
			m_uv_dir[i][1] = factor / div;
		}
	}
}

void face::compute_interpolation_weights(const e3ga::vector &pt, e3ga::mv::Float iw[3], const vertex &vtx0) const {
	// first compute the interpolation factors f[0], f[1], f[2]
	e3ga::mv::Float u_d, v_d, w_d;
	e3ga::vector d(_vector(pt - vtx0.get_pt()));
	u_d = _Float(m_edgeDir[0] % d);
	v_d = _Float(m_edgeDir[1] % d);
	w_d = _Float(m_edgeDir[2] % d);

	iw[0] = (m_v2 * u_d - m_u_v * v_d) / m_u_w2 + 1.0;
	iw[1] = (m_w2 * v_d - m_v_w * w_d) / m_u_w2;
	iw[2] = (m_u2 * w_d - m_w_u * u_d) / m_u_w2;
}

bool face::find_intersection(const e3ga::vector &pt, const e3ga::vector &dir, surface_point &spt, bool find_closest) const {
	// compute how far to 'travel' along the line to get to the plane
	e3ga::mv::Float sd = _Float((pt % -dual(m_pl)) - m_d);
	e3ga::mv::Float d = -sd / _Float(-dual(m_pl) % dir);

	// check if it is nearer to the starting point of the line than the previous
	if (d < RAY_INTERSECTION_MIN_DISTANCE)
		return false;

	// compute where the line intersects the plane
	e3ga::vector ptLocal = _vector(pt + d * dir);

	// compute if this intersection point is inside the triangle
	if (_Float((-dual(m_edgePl[0]) % ptLocal) - m_edgeD[0]) < 0.0) return false;
	if (_Float((-dual(m_edgePl[1]) % ptLocal) - m_edgeD[1]) < 0.0) return false;
	if (_Float((-dual(m_edgePl[2]) % ptLocal) - m_edgeD[2]) < 0.0) return false;

	// if no precise info required, we're done:
	if (!find_closest) {
		spt.set_valid(true);
		spt.set_distance_local(d);
		return true;
	}
	else {
		if ((!spt.get_valid()) ||
			(spt.get_valid() && (d < spt.get_distance_local()))) {
			spt.set_valid(true);
			spt.set_distance_local(d);
			spt.set_pt_local(ptLocal);
			spt.set_side((sd < 0) ? -1 : 1);
			return true;
		}
		else return false;
	}
}

void face::compute_surface_point_details(surface_point &spt, const model &m) const {
	// get the three vertices of this face
	const vertex &vtx0(get_vtx(0, *m.get_mesh()));
	const vertex &vtx1(get_vtx(1, *m.get_mesh()));
	const vertex &vtx2(get_vtx(2, *m.get_mesh()));

	// compute the interpolation weights
	e3ga::mv::Float iw[3]; // iw = interpolation weights
	compute_interpolation_weights(spt.get_pt_local(), iw, vtx0);

	// compute the interpolated texture point
	e3ga::mv::Float uv[2] = {
			iw[0] * vtx0.get_tex_uv()[0] + iw[1] * vtx1.get_tex_uv()[0] + iw[2] * vtx2.get_tex_uv()[0],
			iw[0] * vtx0.get_tex_uv()[1] + iw[1] * vtx1.get_tex_uv()[1] + iw[2] * vtx2.get_tex_uv()[1]
	};
	spt.set_tex_uv(uv);

	// compute interpolated, normalized surface normal
	e3ga::bivector pl = e3ga::_bivector(unit_e(iw[0] * vtx0.get_pl() + iw[1] * vtx1.get_pl() + iw[2] * vtx2.get_pl()));

	// apply bumpmap if it is enabled:
	// todo: clean up GA usage in here . . .
	if (m.get_bump_map_texture().enabled()) {
		float bump[3];
		m.get_bump_map_texture().get_bump_map_value(spt.get_tex_uv(), bump);

		// transform bumpmap from the 2D 'bumpmap frame' to the local 3D frame
		e3ga::bivector bump_pl = e3ga::_bivector(dual(m.get_bumpyness() * 
			((m_uv_dir[0][1] * bump[0] + m_uv_dir[1][1] * bump[1]) * m_edgeDir[2] -
			(m_uv_dir[0][0] * bump[0] + m_uv_dir[1][0] * bump[1]) * m_edgeDir[0])));

		pl = unit_e(pl + bump_pl);
	}

	// set surface attitude in model frame
	spt.set_pl_local(pl);
}
