
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <string.h>
#include <algorithm>

#include "mesh.h"
#include "bsp_node.h"
#include "c3ga_util.h"
#include "c3ga.h"
#include "surface_point.h"
#include "ray.h"

int intersectLineSphere(const e3ga::vector &pt, const e3ga::vector &dir, const e3ga::vector &spherePt, e3ga::mv::Float sphereSize) {
	// compute the point closest  to the sphere center on the line
	return _float(norm_e2((pt + ((spherePt - pt) % dir) * dir) - spherePt)) < (sphereSize * sphereSize);
}

int intersectLineSpherePoints(const e3ga::vector &pt, const  e3ga::vector &dir, const  e3ga::vector &spherePt, e3ga::mv::Float sphereSize,  e3ga::vector isPt[2]) {
	e3ga::vector clPt = _vector(pt + ((spherePt - pt) % dir) * dir); // closest point
	e3ga::mv::Float nd2 = _float(norm_e2(clPt - spherePt)) / (sphereSize * sphereSize); // normalize distance squared
	if (nd2 >= 1.0) return 0; // no intersection
	e3ga::mv::Float f = sphereSize * sqrt(1 - nd2); // _Float
	// compute intersection points:
	isPt[0] = _vector(clPt + dir * f);
	isPt[1] = _vector(clPt - dir * f);
	return 1;
}


mesh::mesh() {
	m_bsp_node = NULL;
}

mesh::mesh(const mesh &m) : m_vertex(m.m_vertex),
		m_face(m.m_face), 
		m_bounding_sphere_center(m.m_bounding_sphere_center),
		m_bounding_sphere_radius(m.m_bounding_sphere_radius),
		m_bsp_node((m.m_bsp_node) ? new bsp_node(*m.m_bsp_node) : NULL),
		m_filename(m.m_filename) {
}

mesh::~mesh() {
	if (m_bsp_node) delete m_bsp_node;
}


mesh &mesh::operator=(const mesh &m) {
	if (this != &m) {
		m_vertex = m.m_vertex;
		m_face = m.m_face;
		m_bounding_sphere_center = m.m_bounding_sphere_center;
		m_bounding_sphere_radius = m.m_bounding_sphere_radius;
		m_filename = m.m_filename;

		m_bsp_node = (m.m_bsp_node) ? new bsp_node(*m.m_bsp_node) : NULL;
	}
	return *this;
}

void mesh::set_vertex(int idx, const vertex &v) {
	if ((idx < 0) || (idx >= (int)m_vertex.size()))
		throw std::string("mesh::set_vertex(): index out of range");
	m_vertex[idx] = v;
}

void mesh::set_face(int idx, const face &f) {
	if ((idx < 0) || (idx >= (int)m_face.size()))
		throw std::string("mesh::set_face(): index out of range");
	m_face[idx] = f;
}

void mesh::set_filename(const std::string &s) {
	m_filename = s;
}	


void mesh::init() {
	init_faces();
	compute_bounding_sphere();
	construct_bsp_tree();
	compute_local_texture_basis();
}

void mesh::init_faces() {
	unsigned int i;

	// compute the constants:
	for (i = 0; i < m_face.size(); i++)
		m_face[i].precompute_constants(*this);

	// remove faces that have no surface area (these occur sometimes in certain models)
	for (i = 0; i < m_face.size(); i++)
		if (!m_face[i].get_valid()) {
			/*
			m_face[i] has no surface area:
			Move last face of the vector to this position,
			downsize the vector, and decrease 'i' by one
			so m_face[i] get's processed again
			*/
			m_face[i] = m_face[m_face.size()-1];
			m_face.resize(m_face.size()-1);
			i--;
		}

}

void mesh::compute_bounding_sphere() {
	/*
	construct a reasonable sphere which surrounds all vertices.
	Steps:
	compute reasonable center point of sphere;
	compute radius of sphere
	*/

	// construct three dual planes, to compute the signed distance to
	int i, j;
	e3ga::vector dir[3];
	e3ga::mv::Float minPos[3], maxPos[3], p, maxD;
	for (i = 0; i < 3; i++) {
		minPos[i] = 10e10f; maxPos[i] = -10e10f; // set to some large value
		dir[i].set(e3ga::vector_e1_e2_e3, (i == 0) ? 1.0 : 0.0, (i == 1) ? 1.0 : 0.0, (i == 2) ? 1.0 : 0.0);
	}

	// compute the minimal and maximal position along each direction
	for (j = 0; j < (int)m_vertex.size(); j++) {
		for (i = 0; i < 3; i++) {
			// compute distance of plane to vertex
			p = _Float(dir[i] % m_vertex[j].get_pt());

			if (p < minPos[i]) minPos[i] = p;
			else if (p > maxPos[i]) maxPos[i] = p;
		}
	}
	// construct reasonable center point of sphere;
	// here we assume the arbitrary initial center was (0.0, 0.0, 0.0)
	// We construct the center point at
	// 0.5 * (minPos[0] + maxPos[0]) * e1 +
	// 0.5 * (minPos[1] + maxPos[1]) * e2 +
	// 0.5 * (minPos[2] + maxPos[2]) * e3
	m_bounding_sphere_center.set(e3ga::vector_e1_e2_e3, 0.5 * (minPos[0] + maxPos[0]), 0.5 * (minPos[1] + maxPos[1]), 0.5 * (minPos[2] + maxPos[2]));

	// compute the radius of the sphere
	maxD = 0.0;
	for (j = 0; j < (int)m_vertex.size(); j++) {
		p = _Float(norm_e(m_vertex[j].get_pt() - m_bounding_sphere_center));
		if (p > maxD) maxD = p;
	}

	// init bounding sphere a _little_ bit (*1e-5) larger than the computed sphere)
	m_bounding_sphere_radius = maxD * (1.0 + 1e-5);

	// check if all vertices lie inside the sphere
	for (j = 0; j < (int)m_vertex.size(); j++) {
		if ( ( p = _Float(norm_e(m_bounding_sphere_center - m_vertex[j].get_pt()) - m_bounding_sphere_radius)) > 0.0) {
			printf("Error computing bounding sphere!!!! Vertex %d lies outside the sphere!\n", j);
			printf("%e should be %e\n", p, 0.0);
		}
	}
}

namespace { // anonymous namspace

// function used with qsort for sorting bsp_sort_data
// TODO: use C++ std lib sort() template . . .
int construct_bsp_tree_sort(const void *v1, const void *v2) {
	struct bsp_sort_data *d1 = (struct bsp_sort_data *)v1, *d2 = (struct bsp_sort_data *)v2;

	if (d1->signed_distance < d2->signed_distance) return -1;
	else return (d1->signed_distance > d2->signed_distance);
}

} // end of anonymous namspace

void mesh::construct_bsp_tree() {
	int i;

	// get memory used to sort vertices
	std::vector<struct bsp_sort_data> bsp_vertex(m_vertex.size());

	// assign the correct index to each entry in bsp_vertex
	for (i = 0; i < (int)bsp_vertex.size(); i++) 
		bsp_vertex[i].idx = i;

	/*
	construct the bsp tree...

	0 is the 'direction' of the first partition plane 
	16 is the constant which determines how many vertices end up in one partition
	*/
	std::vector<unsigned int> lastSizes(3);
	m_bsp_node	 = construct_bsp_tree(0, (int)bsp_vertex.size(), &(bsp_vertex[0]), 16, lastSizes);

	// insert all faces into the bsp tree:
	for (i =0; i < get_nb_faces(); i++) {
		std::vector<e3ga::vector> vertex;
		vertex.push_back(get_vertex(get_face(i).get_vtx_idx(0)).get_pt());
		vertex.push_back(get_vertex(get_face(i).get_vtx_idx(1)).get_pt());
		vertex.push_back(get_vertex(get_face(i).get_vtx_idx(2)).get_pt());
		m_bsp_node	->insert_face(i, vertex);
	}
}

bsp_node *mesh::construct_bsp_tree(int base_plane_idx, unsigned int nb_vertices, struct bsp_sort_data*vertex, unsigned int desired_vertices_per_partition, std::vector<unsigned int> lastSizes) const {
	lastSizes.erase(lastSizes.begin());
	lastSizes.push_back(nb_vertices);

	// we are done if the number of vertices <= the desired number of vertices per partition
	if ((nb_vertices <= desired_vertices_per_partition) || 
		((lastSizes[0] == lastSizes[1]) && (lastSizes[1] == lastSizes[2])))
		return construct_bsp_leaf(/*planes, */nb_vertices, vertex);

	/*
	Get the 'base plane' we are going to use for partitioning the space:
	The integer 'basePlaneIdx' tells us what plane to use:
	*/
	e3ga::vector base_plane_normal;
	if (base_plane_idx == 0) base_plane_normal = _vector(e3ga::e1);
	else if (base_plane_idx == 1) base_plane_normal = _vector(e3ga::e2);
	else base_plane_normal = _vector(e3ga::e3);

	/*
	Compute the signed distance of each vertex to the plane.
	This distance is stored inside the vertices, 
	so we can use it to sort them:
	*/
	for (unsigned int i = 0; i < nb_vertices; i++) {
		vertex[i].signed_distance = _Float(base_plane_normal % get_vertex(vertex[i].idx).get_pt());
	//	printf("d = %f\n", vertex[i].signed_distance);
	}
	       
	/*
	Use quicksort to sort the vertices.
	The function 'constructBSPTreeSort()' compares the
	signed_distance fields of the vertices.
	Todo: use stl quick sort . . . 
	*/
	qsort(vertex, nb_vertices, sizeof(bsp_sort_data), construct_bsp_tree_sort);

	/*
	The required translation of the base plane is now simply the 
	average of the signed distance of the two vertices that came out 
	in the middle after sorting:
	*/
	int split_idx = nb_vertices / 2;
	e3ga::mv::Float d = 0.5 * (vertex[split_idx-1].signed_distance + vertex[split_idx].signed_distance);

	int upto = split_idx;
	int startat = split_idx;
	/*
	Temp test . . . 
	*/
	if (vertex[split_idx-1].signed_distance == vertex[split_idx].signed_distance) {
		while (((upto+1) < (int)nb_vertices) &&
			(vertex[upto].signed_distance == vertex[upto+1].signed_distance))
			upto++;
		while (((startat-1) >= 0) &&
			(vertex[startat].signed_distance == vertex[startat-1].signed_distance))
			startat--;
	}

	/*
	We now translate the dual base plane, (un)dualize it and store
	it in the m_pl of the BSP node:
	*/
//	printf("dual_base_plane = %s\n", dual_base_plane.string());
	//c3ga pl = -(dual_base_plane + d * c3ga::ni).dual(); 
//	printf("pl = %s\n", pl.string ());

	
	// check if partition plane was OK:
	for (int i = 0; i < split_idx; i++) {
		e3ga::mv::Float d2 = _Float(base_plane_normal % get_vertex(vertex[i].idx).get_pt()) - d;
		if (d2 > 0.0)
			printf("mesh::construct_bsp_tree(): error: d2- = %f!!!!!!!!!!!!!!\n", d2);
	}
	for (unsigned int i = split_idx; i < nb_vertices; i++) {
		e3ga::mv::Float d2 = _Float(base_plane_normal % get_vertex(vertex[i].idx).get_pt()) - d;
		if (d2 < 0.0)
			printf("mesh::construct_bsp_tree(): error: d2+ = %f!!!!!!!!!!!!!!\n", d2);
	}

	struct bsp_sort_data *vertex2 = vertex;
	std::vector<struct bsp_sort_data> bsp_vertex2;
	if (upto != startat) {
		bsp_vertex2.resize(nb_vertices);
		vertex2 = &(bsp_vertex2[0]);
		memcpy(vertex2, vertex, sizeof(struct bsp_sort_data) * nb_vertices);
	}

	return new bsp_node(_bivector(dual(base_plane_normal)), d, 
		construct_bsp_tree((base_plane_idx+1)%3, upto, vertex, desired_vertices_per_partition, lastSizes),
		construct_bsp_tree((base_plane_idx+1)%3, nb_vertices - startat, vertex2 + startat, desired_vertices_per_partition, lastSizes));
}


bsp_node *mesh::construct_bsp_leaf(unsigned int nb_vertices, struct bsp_sort_data*vertex) const {
	return new bsp_node();
}

void mesh::compute_local_texture_basis() {
  	for (unsigned int i = 0; i < m_face.size(); i++) {
		m_face[i].compute_local_texture_basis(*this);
	}
}

bool mesh::find_intersection(const ray &R, surface_point &spt, bool find_closest) const {
//	c3ga ray_dual_plane = (c3ga::no << R.get_position()) << R.get_direction();


	// The point must be either inside the sphere . . . 
	if (_Float(norm_e(R.get_position() - get_bounding_sphere_center())) > get_bounding_sphere_radius()) {
		/// . . . or the sphere must be ahead of the ray:
//		double x = (R_dual_plane % get_bounding_dual_sphere()).scalar();
		if (_Float(R.get_direction() % (get_bounding_sphere_center() - R.get_position())) < 0.0)
			return false;
	}

	// OK, the sphere is ahead of the ray, now check for intersection
	if (!intersectLineSphere(R.get_position(), R.get_direction(), get_bounding_sphere_center(), get_bounding_sphere_radius()))
		return false;

	e3ga::vector isPt[2];
	intersectLineSpherePoints(R.get_position(), R.get_direction(), get_bounding_sphere_center(), get_bounding_sphere_radius(), isPt);

	return find_intersection_bsp(isPt[0], isPt[1], R, spt, find_closest);
}

bool mesh::find_intersection_bsp(const e3ga::vector &start_pt, const e3ga::vector &end_pt, const ray &R, surface_point &spt, bool find_closest) const {
	// determine where those point are relative to (the start of) the ray
	e3ga::mv::Float d1, d2;
	d1 = _Float((start_pt - R.get_position()) % R.get_direction());
	d2 = _Float((end_pt - R.get_position()) % R.get_direction());

	// descent down the bsp tree (but first order fpt1 and fpt2)
	if (d1 < d2) {
		bsp_descent_info bdi(this, start_pt, d1, end_pt, d2, R, spt, find_closest);
		return get_bsp_node()->find_intersection(bdi);
	}
	else {
		bsp_descent_info bdi(this, end_pt, d2, start_pt, d1, R, spt, find_closest);
		return get_bsp_node()->find_intersection(bdi);
	}
}

